import importlib
import inspect
import logging
import os
import subprocess
import time
from typing import Optional

import typer
from rich.console import Console
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn
from rich.prompt import Confirm, Prompt
from typing_extensions import Annotated

from .pharia_skill_cli import PhariaSkillCli, Registry

logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    handlers=[logging.StreamHandler()],
)

logger = logging.getLogger(__name__)
console = Console()


def setup_wasi_deps() -> None:
    """Download the Pydantic WASI wheels if they are not already present."""
    PYDANTIC_CORE_VERSION = "2.27.2"
    WASI_DEPS_PATH = "wasi_deps"
    if os.path.exists(WASI_DEPS_PATH):
        if not os.path.exists(
            f"{WASI_DEPS_PATH}/pydantic_core-{PYDANTIC_CORE_VERSION}.dist-info"
        ):
            logger.info("Deleting outdated Pydantic Wasi wheels...")
            subprocess.run(["rm", "-rf", WASI_DEPS_PATH])

    if not os.path.exists(WASI_DEPS_PATH):
        logger.info("Downloading Pydantic Wasi wheels...")
        subprocess.run(
            [
                "pip3",
                "install",
                "--target",
                WASI_DEPS_PATH,
                "--only-binary",
                ":all:",
                "--platform",
                "any",
                "--platform",
                "wasi_0_0_0_wasm32",
                "--python-version",
                "3.12",
                "--index-url",
                "https://benbrandt.github.io/wasi-wheels/",
                "--extra-index-url",
                "https://pypi.org/simple",
                f"pydantic-core=={PYDANTIC_CORE_VERSION}",
            ],
            check=True,
        )


class BuildError(Exception):
    def __init__(self, message: str):
        self.message = message


class ModuleError(Exception):
    pass


def inspect_wit_world(module_path: str) -> str:
    """Determine the world that a Skill should be build against.

    The SDK supports multiple wit worlds (e.g. `skill` and `message-stream-skill`).
    Each decorator targets a particular world.

    This function inspects a module, looking for one of the exported classes
    to determine which world the Skill should be build against.
    """
    if contains_class(module_path, "SkillHandler"):
        return "skill"
    elif contains_class(module_path, "MessageStream"):
        return "message-stream-skill"
    else:
        raise ValueError(
            f"Unable to find a Skill in {module_path}. "
            "Did you add the @skill or @message_stream decorator to the Skill function?"
        )


def contains_class(module_path: str, class_name: str) -> bool:
    """Check if a class named `class_name` exists in the given module."""
    module = importlib.import_module(module_path)
    return any(
        name == class_name and inspect.isclass(value)
        for name, value in vars(module).items()
    )


def run_componentize_py(skill_module: str, output_file: str, unstable: bool) -> str:
    """Build the skill to a WASM component using componentize-py.

    The call to componentize-py targets the `skill` wit world and adds the downloaded
    Pydantic WASI wheels to the Python path.

    Returns:
        str: The path to the generated WASM file.
    """
    setup_wasi_deps()
    args = ["--all-features"] if unstable else []
    world = inspect_wit_world(skill_module)
    command = [
        "componentize-py",
        *args,
        "-w",
        world,
        "componentize",
        skill_module,
        "-o",
        output_file,
        "-p",
        ".",
        "-p",
        "wasi_deps",
    ]

    try:
        subprocess.run(
            command,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
    except subprocess.CalledProcessError as e:
        raise BuildError(e.stderr)

    return output_file


def display_publish_suggestion(wasm_file: str) -> None:
    """Display a colorful suggestion to publish the skill.

    Args:
        wasm_file: Path to the generated WASM file.
    """
    wasm_filename = wasm_file.lstrip("./")
    publish_command = f"pharia-skill publish {wasm_filename} --tag [TAG] --name [NAME]"

    console.print(
        Panel.fit(
            f"[bold]Skill:[/bold] [cyan]{wasm_filename}[/cyan]\n\n"
            f"[yellow]To publish, run:[/yellow]\n"
            f"[cyan]{publish_command}[/cyan]",
            title="[bold green]Build Successful[/bold green]",
            border_style="green",
            padding=(1, 2),
        )
    )


def publish_skill(skill_path: str, name: Optional[str], tag: str) -> None:
    """Publish a skill with progress indicator and success message.

    Args:
        skill_path: Path to the WASM file to publish.
        name: Name to publish the skill as, or None to use the filename.
        tag: Tag to publish the skill with.
    """
    if not skill_path.endswith(".wasm"):
        skill_path += ".wasm"

    display_name = name if name else skill_path.replace(".wasm", "")

    cli = PhariaSkillCli()

    try:
        registry = Registry.from_env()
    except KeyError as e:
        console.print(
            Panel(
                f"The environment variable [yellow]{e}[/yellow] is not set.",
                title="[bold red]Error[/bold red]",
                border_style="red",
                padding=(1, 1),
            )
        )
        raise typer.Exit(code=1)

    start_time = time.time()
    with Progress(
        SpinnerColumn(),
        TextColumn(f"Publishing [cyan]{display_name}[/cyan]:[cyan]{tag}[/cyan]..."),
        TimeElapsedColumn(),
        console=console,
        transient=True,
    ) as progress:
        task = progress.add_task("", total=None)
        cli.publish(skill_path, name, tag, registry)
        progress.update(task, completed=True)

    console.print(
        Panel.fit(
            f"[bold]Skill:[/bold] [cyan]{display_name}[/cyan]\n"
            f"[bold]Tag:[/bold] [cyan]{tag}[/cyan]\n"
            f"[bold]Registry:[/bold] [cyan]{registry.registry}/{registry.repository}[/cyan]\n\n"
            f"Published in [yellow]{time.time() - start_time:.2f}[/yellow] seconds",
            title="[bold green]Publish Successful[/bold green]",
            border_style="green",
            padding=(1, 2),
        )
    )


def prompt_for_publish(wasm_file: str) -> None:
    """Prompt the user to publish the skill.

    Args:
        wasm_file: Path to the generated WASM file.
    """
    wasm_filename = wasm_file.lstrip("./")

    # Ask if the user wants to publish now with a more engaging prompt
    if Confirm.ask(
        "\n[bold yellow]Would you like to publish this skill now?[/bold yellow]",
        default=True,
        console=console,
    ):
        tag = Prompt.ask(
            "[bold cyan]Enter tag[/bold cyan]",
            default="latest",
            show_default=True,
            console=console,
        )

        name_default = wasm_filename.replace(".wasm", "")
        name = Prompt.ask(
            "[bold cyan]Enter name[/bold cyan]",
            default=name_default,
            show_default=True,
            console=console,
        )

        # Publish the skill
        publish_skill(wasm_filename, name, tag)


app = typer.Typer(rich_markup_mode="rich")


@app.callback()
def callback() -> None:
    """
    [bold green]Pharia Skill CLI Tool[/bold green].

    A tool for building and publishing Pharia Skills.
    """


@app.command()
def build(
    skill: Annotated[
        str,
        typer.Argument(help="Python module of the skill to build", show_default=False),
    ],
    unstable: Annotated[
        bool,
        typer.Option(
            help="Enable unstable features for testing. Don't try this at home."
        ),
    ] = False,
    interactive: Annotated[
        bool,
        typer.Option(
            help="Prompt for publishing after building.",
            show_default=True,
        ),
    ] = True,
) -> None:
    """
    [bold blue]Build[/bold blue] a skill.

    Compiles a Python module into a WebAssembly component.
    """
    if "/" in skill or skill.endswith(".py"):
        suggestion = skill
        if skill.endswith(".py"):
            suggestion = skill[:-3]
        if "/" in suggestion:
            suggestion = suggestion.replace("/", ".")

        console.print(
            Panel(
                f"Argument must be a fully qualified Python module name, not [cyan]{skill}[/cyan]\n\n"
                f"[yellow]Did you mean?[/yellow] [green]{suggestion}[/green]\n\n"
                f"[italic]Example: Use [green]my_package.my_module[/green] instead of [red]my_package/my_module.py[/red][/italic]",
                title="[bold red]Error[/bold red]",
                border_style="red",
                padding=(1, 1),
            )
        )
        raise typer.Exit(code=1)

    output_file = f"./{skill.split('.')[-1]}.wasm"
    with Progress(
        SpinnerColumn(),
        TextColumn(
            f"Building WASM component [cyan]{output_file}[/cyan] from module [cyan]{skill}[/cyan]..."
        ),
        TimeElapsedColumn(),
        console=console,
        transient=True,
    ) as progress:
        task = progress.add_task("", total=None)
        wasm_file = run_componentize_py(skill, output_file, unstable)
        progress.update(task, completed=True)

    if wasm_file and interactive:
        display_publish_suggestion(wasm_file)
        prompt_for_publish(wasm_file)


@app.command()
def publish(
    skill: Annotated[
        str,
        typer.Argument(
            help="A path to a Wasm file containing a Skill.", show_default=False
        ),
    ],
    name: Annotated[
        Optional[str],
        typer.Option(
            help="The name to publish the Skill as. If not provided, it is inferred based on the Wasm filename.",
            show_default="The filename",
        ),
    ] = None,
    tag: Annotated[str, typer.Option(help="An identifier for the Skill.")] = "latest",
) -> None:
    """
    [bold blue]Publish[/bold blue] a skill.

    Publishes a WebAssembly component to the Pharia Skill registry.
    """
    publish_skill(skill, name, tag)


if __name__ == "__main__":
    app()



import datetime as dt
from dataclasses import asdict, dataclass, field
from typing import Any, Literal

from pydantic import model_serializer


@dataclass
class DocumentPath:
    """Path identifying a document.

    A DocumentPath consists of a namespace, within the namespace a collection and within the collection a document has a name.

    Attributes:
        namespace (str): The namespace.
        collection (str): The collection within the namespace.
        name (str): The name identifying the document in the collection.
    """

    namespace: str
    collection: str
    name: str


@dataclass
class IndexPath:
    """Which documents you want to search in, and which type of index should be used.

    Attributes:
        namespace (string): The namespace the collection belongs to.
        collection (string): The collection you want to search in.
        index (str): The search index you want to use for the collection.
    """

    namespace: str
    collection: str
    index: str


@dataclass
class GreaterThan:
    __json_name__ = "greater_than"

    value: float


@dataclass
class GreaterThanOrEqualTo:
    __json_name__ = "greater_than_or_equal_to"

    value: float


@dataclass
class LessThan:
    __json_name__ = "less_than"

    value: float


@dataclass
class LessThanOrEqualTo:
    __json_name__ = "less_than_or_equal_to"

    value: float


@dataclass
class After:
    __json_name__ = "after"

    value: dt.datetime

    def __post_init__(self) -> None:
        assert self.value.tzinfo is not None, "Datetimes must be timezone-aware"


@dataclass
class AtOrAfter:
    __json_name__ = "at_or_after"

    value: dt.datetime

    def __post_init__(self) -> None:
        assert self.value.tzinfo is not None, "Datetimes must be timezone-aware"


@dataclass
class Before:
    __json_name__ = "before"

    value: dt.datetime

    def __post_init__(self) -> None:
        assert self.value.tzinfo is not None, "Datetimes must be timezone-aware"


@dataclass
class AtOrBefore:
    __json_name__ = "at_or_before"

    value: dt.datetime

    def __post_init__(self) -> None:
        assert self.value.tzinfo is not None, "Datetimes must be timezone-aware"


@dataclass
class EqualTo:
    __json_name__ = "equal_to"

    value: str | int | bool


@dataclass
class IsNull:
    __json_name__ = "is_null"

    value: Literal[True] = True


"""This condition matches all metadata fields with a value of null."""


FilterCondition = (
    GreaterThan
    | GreaterThanOrEqualTo
    | LessThan
    | LessThanOrEqualTo
    | After
    | AtOrAfter
    | Before
    | AtOrBefore
    | EqualTo
    | IsNull
)


@dataclass
class MetadataFilter:
    """Matches sections whose metadata fields match the given condition. You must specify the field, and can only specify a single condition.

    While the Document Index also offers a `Modality` filter, we do not expose this to the developer.
    The reasoning is that we only allow for text modalities in the Kernel. So for each search request,
    we append a `Modality` filter that only allows for text modalities.

    Attributes:
        field (str): The metadata field on which to filter search results.
            Field names must only contain alphanumeric characters, dashes and underscores.
            Nested fields can be specified using dot notation (e.g. 'a.b').
            Array-valued fields can either use a wildcard specifier (e.g. 'a[].b') or a specific index (e.g. 'a[1].b').
            The maximum length of the field name is 1000 characters.
        condition (FilterCondition): The condition to filter on.
    """

    field: str
    condition: FilterCondition

    def serialize(self) -> dict[str, Any]:
        """How to serialize a metadata filter to a dictionary.

        It would be nice to specify this as a `model_serializer` and let pydantic handle
        the serialization. However, as we are already doing custom serialization on the
        outside, and this is not a Pydantic model we could call `.model_dump()` on, it
        seems to be the simplest solution to just implement the serialization manually.
        """

        return {
            "metadata": {
                "field": self.field,
                self.condition.__json_name__: self.condition.value,
            }
        }


@dataclass
class Without:
    """Logical conjunction of negations, i.e. forms the predicate "(NOT filterCondition1) AND (NOT filterCondition2) AND ..."

    Attributes:
        value (list[Filter]): The list of filter conditions.
    """

    value: list[MetadataFilter]

    def serialize(self) -> dict[str, list[Any]]:
        return {"without": [filter.serialize() for filter in self.value]}


@dataclass
class WithOneOf:
    """Logical disjunction, i.e. forms the predicate "filterCondition1 OR filterCondition2 OR ..."

    Attributes:
        value (list[Filter]): The list of filter conditions.
    """

    value: list[MetadataFilter]

    def serialize(self) -> dict[str, list[Any]]:
        return {"with_one_of": [filter.serialize() for filter in self.value]}


@dataclass
class With:
    """Logical conjunction, i.e. forms the predicate "filterCondition1 AND filterCondition2 AND ..."

    Attributes:
        value (list[Filter]): The list of filter conditions.
    """

    value: list[MetadataFilter]

    def serialize(self) -> dict[str, list[Any]]:
        return {"with": [filter.serialize() for filter in self.value]}


SearchFilter = Without | With | WithOneOf
"""A logical combination of filter conditions."""


@dataclass
class SearchRequest:
    """A request to search the document index.

    Attributes:
        index_path (IndexPath): The index path to search in.
        query (str): The query to search for.
        max_results (int): Maximum number of results to return. Defaults to 1.
        min_score (float | None): Filter out results with a cosine similarity score below this value.
            Scores range from -1 to 1. For searches on hybrid indexes, the Document Index applies the min_score to the semantic results before fusion of result sets.
            As fusion re-scores results, returned scores may exceed this value.
        filters (list[SearchFilter]): A filter for search results that restricts the results to those document sections that match the filter criteria.
            The individual conditions of this array are AND-combined (i.e. all conditions must match).
            This can for example be used to restrict the returned sections based on their modality (i.e. image or text), or on their metadata.
    """

    index_path: IndexPath
    query: str
    max_results: int = 1
    min_score: float | None = None
    filters: list[SearchFilter] = field(default_factory=list)

    @model_serializer()
    def serialize(self) -> dict[str, Any]:
        return {
            "index_path": asdict(self.index_path),
            "query": self.query,
            "max_results": self.max_results,
            "min_score": self.min_score,
            "filters": [filter.serialize() for filter in self.filters],
        }


@dataclass
class Cursor:
    """A position within a document.

    The cursor is always inclusive of the current position, in both start and end positions.

    Attributes:
        item (int): Index of the item in the document. A document is an array of text and image elements. These elements are referred to as items.
        position (int): The character position the cursor can be found at within the string.
    """

    item: int
    position: int


@dataclass
class SearchResult:
    """The relevant documents as result of a search request.

    Attributes:
        document_path (DocumentPath): The path to a document. A path uniquely identifies a document among all managed documents.
        content (str): The text of the found section. As we do not support multi-modal, this is always a string.
        score (float): Search score of the found section, where a higher score indicates a closer match.
            Will be between -1 and 1. A score closer to -1 indicates the section opposes the query.
            A score close 0 suggests the section is unrelated to the query.
            A score close to 1 suggests the section is related to the query.
        start (Cursor): Where the result starts in the document.
        end (Cursor): Where the result ends in the document.
    """

    document_path: DocumentPath
    content: str
    score: float
    start: Cursor
    end: Cursor


@dataclass
class Text:
    """A text section that is part of a document.

    If the document only contains text, then the contents of the document is a list of
    length one, where the only element is a `Text`."""

    text: str
    modality: Literal["text"] = "text"


@dataclass
class Image:
    """An image that is part of a document.

    At the moment, we do not expose the image contents, as none of the models
    support multi-modal inputs. We still inform the developer that the document
    contains an image.
    """

    modality: Literal["image"] = "image"


Modality = Text | Image
"""A document is made up of subsections of different modalities.

For example, if a document is a long text with an image in the middle, then
it will be represented by a list of length three, where the first and last
elements are `Text` and the middle element is `Image`.
"""


JsonSerializable = dict[str, Any] | list[Any] | str | int | float | bool | None
"""Represent any value that can be serialized/deserialized to/from JSON.

Used to represent the return type of `document_metadata` which is any valid JSON value.
"""


@dataclass
class Document:
    """A document in the Document Index.

    Attributes:
        path (DocumentPath): The path that identifies the document.
        contents (list[Modality]): The contents of the document. Split into sections of different modalities.
        metadata (JsonSerializable): The (custom) metadata of the document.
    """

    path: DocumentPath
    contents: list[Modality]
    metadata: JsonSerializable

    @property
    def text(self) -> str:
        """Concatenate the text contents of the document."""
        return "\n\n".join(
            text.text for text in self.contents if isinstance(text, Text)
        )



from pharia_skill.message_stream.writer import MessageItem, MessageWriter, Payload


class MessageRecorder(MessageWriter[Payload]):
    """A message writer that can be passed into a `message_stream` skill at testing time.

    It allows to inspect the output that a skill produces.

    Example::

        from pharia_skill import Csi, message_stream, MessageAppend, MessageBegin, MessageEnd
        from pharia_skill.testing import MessageWriter, MessageRecorder

        @message_stream
        def my_skill(csi: Csi, writer: MessageWriter, input: Input) -> None:
            ...

        def test_my_skill():
            csi = DevCsi()
            writer = MessageRecorder()
            input = Input(topic="The meaning of life")
            my_skill(csi, writer, input)
            assert writer.items == [
                MessageBegin(role="assistant"),
                MessageAppend(text="The meaning of life"),
                MessageEnd(payload=None),
            ]
    """

    def __init__(self) -> None:
        self.items: list[MessageItem[Payload]] = []

    def write(self, item: MessageItem[Payload]) -> None:
        self.items.append(item)



from collections.abc import Callable
from dataclasses import dataclass
from typing import Generic, Protocol, TypeVar

from pydantic import BaseModel

from pharia_skill.csi.inference import ChatStreamResponse, CompletionStreamResponse

Payload = TypeVar("Payload", bound=BaseModel | None)


@dataclass
class MessageBegin:
    role: str | None


@dataclass
class MessageAppend:
    text: str


@dataclass
class MessageEnd(Generic[Payload]):
    payload: Payload | None


MessageItem = MessageBegin | MessageAppend | MessageEnd[Payload]


class MessageWriter(Protocol, Generic[Payload]):
    """Write messages to the output stream."""

    def write(self, item: MessageItem[Payload]) -> None: ...

    def begin_message(self, role: str | None = None) -> None:
        self.write(MessageBegin(role))

    def append_to_message(self, text: str) -> None:
        self.write(MessageAppend(text))

    def end_message(self, payload: Payload | None = None) -> None:
        self.write(MessageEnd(payload))

    R = TypeVar("R", CompletionStreamResponse, ChatStreamResponse)

    def forward_response(
        self, response: R, payload: Callable[[R], Payload] | Payload | None = None
    ) -> None:
        match response:
            case CompletionStreamResponse():
                self._forward_completion(response)
            case ChatStreamResponse():
                self._forward_chat(response)

    def _forward_completion(
        self,
        response: CompletionStreamResponse,
        payload: Callable[[CompletionStreamResponse], Payload] | Payload | None = None,
    ) -> None:
        self.begin_message()
        for append in response.stream():
            self.append_to_message(append.text)
        self.end_message(payload(response) if callable(payload) else payload)

    def _forward_chat(
        self,
        response: ChatStreamResponse,
        payload: Callable[[ChatStreamResponse], Payload] | Payload | None = None,
    ) -> None:
        self.begin_message(response.role)
        for append in response.stream():
            self.append_to_message(append.content)
        self.end_message(payload(response) if callable(payload) else payload)



from .client import StudioClient
from .exporter import SpanClient, StudioExporter, StudioSpanProcessor
from .span import StudioSpan

__all__ = [
    "StudioClient",
    "StudioExporter",
    "StudioSpan",
    "SpanClient",
    "StudioSpanProcessor",
]



"""
Two implementations of the Cognitive System Interface (CSI) that can be used for testing and development.

When Skills are run in the Kernel, the CSI is provided via an Application Binary Interface.
This interface is defined via the WASM Interface Type (WIT) language.
For development and debugging, Skills can also run in a local Python environment.
The CSI which is available to the Skill at runtime can be substituted with a DevCSI which is backed by HTTP requests against a running instance of the Kernel.
Developers can write tests, step through their Python code and inspect the state of variables.
"""

from .dev import DevCsi, MessageRecorder
from .stub import StubCsi

__all__ = ["StubCsi", "DevCsi", "MessageRecorder"]
